package com.mdtexfx.processors;

import com.mdtexfx.processors.highlighters.SyntaxHighlighter;
import com.vladsch.flexmark.ast.*;
import com.vladsch.flexmark.html.HtmlRenderer;
import com.vladsch.flexmark.parser.Parser;
import com.vladsch.flexmark.util.ast.*;

/**
 * Responsible for transforming Markdown into HTML.
 */
@SuppressWarnings("SameParameterValue")
public class MarkdownProcessor extends ExecutorProcessor<String> {
  private final IParse mParser = Parser.builder().build();
  private final IRender mRenderer = HtmlRenderer.builder().build();
  private final SyntaxHighlighter mSyntax;
  private final NodeVisitor mVisitor;

  /**
   * Constructs a new processor that converts Markdown to HTML.
   *
   * @param successor The {@link Processor} that will handle the HTML output by
   *                  this {@link Processor}.
   * @param context   Contains the {@link SyntaxHighlighter} used to delegate
   *                  syntax highlight requests to the text editor.
   */
  public MarkdownProcessor(
      final Processor<String> successor, final ProcessorContext context ) {
    super( successor );
    mSyntax = context.getSyntaxHighlighter();
    mVisitor = new NodeVisitor(
        create( Text.class, node -> container( node, "text" ) ),
        create( Heading.class, node -> heading( node, "h" + node.getLevel() ) ),
        create( Code.class, node -> delimited( node, "code" ) ),
        create( Emphasis.class, node -> delimited( node, "emphasis" ) ),
        create( StrongEmphasis.class, node -> delimited( node, "strong" ) ),
        create( FencedCodeBlock.class, node -> fenced( node, "pre" ) ),
        create( BlockQuote.class, node -> container( node, "blockquote" ) ),
        create( BulletListItem.class, node -> itemized( node, "bulleted" ) ),
        create( OrderedListItem.class, node -> itemized( node, "enumerated" ) ),
        create( Link.class, node -> link( node, "link" ) ),
        create( HtmlEntity.class, node -> container( node, "entity" ) )
    );
  }

  @Override
  public String apply( final String markdown ) {
    final var root = mParser.parse( markdown );
    mVisitor.visit( root );
    return mRenderer.render( root );
  }

  private <N extends Node> VisitHandler<N> create(
      final Class<N> clazz, final Visitor<N> visitor ) {
    return new VisitHandler<>( clazz, visitor );
  }

  private void highlight( final int start, final int end, final String style ) {
    assert mSyntax != null;
    mSyntax.highlight( start, end, style );
  }

  private void container( final Node node, final String style ) {
    highlight( node.getStartOffset(), node.getEndOffset(), style );
  }

  private void heading( final Heading node, final String style ) {
    highlight(
        node.getOpeningMarker().getStartOffset(),
        node.getText().getEndOffset(),
        style
    );
  }

  private void delimited( final DelimitedNode node, final String style ) {
    highlight(
        node.getOpeningMarker().getStartOffset(),
        node.getClosingMarker().getEndOffset(),
        style
    );
  }

  private void fenced( final FencedCodeBlock node, final String style ) {
    if( node.getClosingFence().getEndOffset() != 0 ) {
      highlight(
          node.getOpeningMarker().getStartOffset(),
          node.getClosingMarker().getEndOffset(),
          style
      );
    }
  }

  private void itemized( final ListItem node, final String style ) {
    highlight(
        node.getOpeningMarker().getStartOffset(),
        node.getOpeningMarker().getEndOffset(),
        style
    );
  }

  private void link( final Link node, final String style ) {
    highlight(
        node.getTextOpeningMarker().getStartOffset(),
        node.getLinkClosingMarker().getEndOffset(),
        style
    );
  }
}
